import errno
import json
import logging
import os
from six.moves import queue
import requests
import subprocess
import threading
import time

logger = logging.getLogger(__name__)

# TODO: get rid of this
try:
    import numpy as np
except ImportError:
    pass


class WandBJSONEncoder(json.JSONEncoder):
    """A JSON Encoder that handles some extra types."""

    def default(self, obj):
        # TODO: Some of this is just guessing. Be smarter.
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return np.asscalar(obj)
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        return json.JSONEncoder.default(self, obj)


def json_dumps_safer(obj, **kwargs):
    """Convert obj to json, with some extra encodable types."""
    return json.dumps(obj, cls=WandBJSONEncoder, **kwargs)


def make_json_if_not_number(v):
    """If v is not a basic type convert it to json."""
    if isinstance(v, (float, int)):
        return v
    return json_dumps_safer(v)


def mkdir_exists_ok(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class SafeSubprocess(object):
    def __init__(self, args, env=None, read_output=False):
        self._args = args
        self._env = env
        self._read_output = read_output
        self._stdout = queue.Queue()
        self._stderr = queue.Queue()
        self._popen = None
        self._stdout_thread = None
        self._stderr_thread = None

    def run(self):
        if self._read_output:
            self._popen = subprocess.Popen(
                self._args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=self._env)
            self._stdout_thread = self._spawn_reader_thread(
                self._popen.stdout, self._stdout)
            self._stderr_thread = self._spawn_reader_thread(
                self._popen.stderr, self._stderr)
        else:
            self._popen = subprocess.Popen(self._args, env=self._env)

    def _spawn_reader_thread(self, filelike, out_queue):
        def _reader_thread(filelike, out_queue):
            while True:
                out = filelike.read(64).decode('utf-8')
                if not out:
                    break
                out_queue.put(out)

        threading.Thread(target=_reader_thread,
                         args=(filelike, out_queue)).start()

    def _read(self, rqueue):
        try:
            return rqueue.get(False)
        except queue.Empty:
            return None

    def _read_all(self, rqueue):
        reads = []
        while True:
            got = self._read(rqueue)
            if got is None:
                break
            reads.append(got)
        return reads

    def _read_stdout(self):
        return self._read_all(self._stdout)

    def _read_stderr(self):
        return self._read_all(self._stderr)

    def poll(self):
        if self._read_output:
            return self._popen.poll(), self._read_stdout(), self._read_stderr()
        else:
            return self._popen.poll()


def request_with_retry(func, *args, **kwargs):
    """Perform a requests http call, retrying with exponetial backoff.

    Args:
        func: An http-requesting function to call, like requests.post
        retries: Maximum retries before giving up.
        *args: passed through to func
        **kwargs: passed through to func
    """
    retries = kwargs.get('retries', 5)
    retry_delay = 2
    retry_count = 0
    while True:
        try:
            response = func(*args, **kwargs)
            response.raise_for_status()
            return True
        except (requests.exceptions.ConnectionError,
                requests.exceptions.HTTPError,
                requests.exceptions.Timeout) as e:
            logger.warning('requests_with_retry encountered retryable exception: %s. args: %s, kwargs: %s',
                           e, args, kwargs)
            if retry_count == retries:
                return e
            retry_count += 1
            time.sleep(retry_delay)
            retry_delay *= 2
        except requests.exceptions.RequestException as e:
            logger.error(
                'requests_with_retry encountered unretryable exception: %s', e)
            return e


def find_runner(program):
    """Return a command that will run program.

    Args:
        program: The string name of the program to try to run.
    Returns:
        string, Runner name.
    """
    if os.path.isfile(program) and not os.access(program, os.X_OK):
        # program is a path to a non-executable file
        try:
            opened = open(program)
        except PermissionError:
            return None
        first_line = opened.readline().strip()
        if first_line.startswith('#!'):
            return first_line[2:]
        if program.endswith('.py'):
            return 'python'
    return None



import operator
import os

import wandb

# Fully implemented here so we don't have to pull in keras as a dependency.
# However, if the user is using this, they necessarily have Keras installed. So we
# could probably selectively build this class only when the user requests it,
# knowing that keras is available.
#
# Or have a separate lib "wandb-keras", then we could use the appropriate Keras
# pieces


class WandBKerasCallback(object):
    """WandB Keras Callback.

    Automatically saves wandb-history.csv and wandb-summary.csv, both tracking
    keras metrics.
    """

    def __init__(self, monitor='val_loss', verbose=0, mode='auto',
                 save_weights_only=False):
        """Constructor.

        Args:
            See keras.ModelCheckpoint for other definitions of other
                arguments.
        """
        self.validation_data = None

        self.monitor = monitor
        self.verbose = verbose
        self.save_weights_only = save_weights_only

        self.filepath = os.path.join(wandb.run.dir, 'model-best.h5')

        # From Keras
        if mode not in ['auto', 'min', 'max']:
            print('WandBKerasCallback mode %s is unknown, '
                  'fallback to auto mode.' % (mode))
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = operator.lt
            self.best = float('inf')
        elif mode == 'max':
            self.monitor_op = operator.gt
            self.best = float('-inf')
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = operator.gt
                self.best = float('-inf')
            else:
                self.monitor_op = operator.lt
                self.best = float('inf')

    def set_params(self, params):
        self.params = params

    def set_model(self, model):
        self.model = model

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        # history
        row = {'epoch': epoch}
        row.update(logs)
        wandb.run.history.add(row)

        # summary
        current = logs.get(self.monitor)
        if current is None:
            print('Can save best model only with %s available, '
                  'skipping.' % (self.monitor))

        if self.monitor_op(current, self.best):
            row.pop('epoch')
            wandb.run.summary.update(row)
            if self.verbose > 0:
                print('Epoch %05d: %s improved from %0.5f to %0.5f,'
                      ' saving model to %s'
                      % (epoch, self.monitor, self.best,
                         current, self.filepath))
            self.best = current
            if self.save_weights_only:
                self.model.save_weights(self.filepath, overwrite=True)
            else:
                self.model.save(self.filepath, overwrite=True)

    def on_batch_begin(self, batch, logs=None):
        pass

    def on_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        pass



import pytest, os, traceback, click
from wandb import cli, Api, GitRepo, __version__
from click.testing import CliRunner
from .api_mocks import *
import netrc, signal, time
import six, time, inquirer, yaml
import git
import webbrowser
import wandb

try:
    # python 3.4+
    from importlib import reload
except ImportError:
    # python 3.2, 3.3
    from imp import reload
except ImportError:
    pass

@pytest.fixture
def runner(monkeypatch):
    monkeypatch.setattr(cli, 'api', Api(default_settings={'project': 'test', 'git_tag': True}, load_settings=False))
    monkeypatch.setattr(click, 'launch', lambda x: 1)
    monkeypatch.setattr(inquirer, 'prompt', lambda x: {'project': 'test_model', 'files': ['weights.h5']})
    monkeypatch.setattr(webbrowser, 'open_new_tab', lambda x: True)
    return CliRunner()

@pytest.fixture
def empty_netrc(monkeypatch):
    class FakeNet(object):
        @property
        def hosts(self):
            return {'api.wandb.ai': None}
    monkeypatch.setattr(netrc, "netrc", lambda *args: FakeNet())

@pytest.fixture
def local_netrc(monkeypatch):
    #TODO: this seems overkill...
    origexpand = os.path.expanduser
    def expand(path):
        return os.path.realpath("netrc") if "netrc" in path else origexpand(path)
    monkeypatch.setattr(os.path, "expanduser", expand)

def git_repo():
    r = git.Repo.init(".")
    open("README", "wb").close()
    r.index.add(["README"])
    r.index.commit("Initial commit")
    return GitRepo(lazy=False)

def test_help(runner):
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    assert 'Weights & Biases' in result.output
    help_result = runner.invoke(cli.cli, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output

def test_version(runner):
    result = runner.invoke(cli.cli, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.output

@pytest.mark.skip(reason='config reworked, fixes coming...')
def test_config(runner, monkeypatch):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.config, ["init"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert "wandb config set" in result.output
        assert os.path.exists("config-defaults.yaml")

@pytest.mark.skip(reason='config reworked, fixes coming...')
def test_config_show(runner, monkeypatch):
    with runner.isolated_filesystem():
        with open("config-defaults.yaml", "w") as f:
            f.write(yaml.dump({'val': {'value': 'awesome', 'desc': 'cool'}, 'bad': {'value':'shit'}}))
        result_py = runner.invoke(cli.config, ["show"])
        result_yml = runner.invoke(cli.config, ["show", "--format", "yaml"])
        result_json = runner.invoke(cli.config, ["show", "--format", "json"])
        print(result_py.output)
        print(result_py.exception)
        print(traceback.print_tb(result_py.exc_info[2]))
        assert "awesome" in result_py.output
        assert "awesome" in result_yml.output
        assert "awesome" in result_json.output

@pytest.mark.skip(reason='config reworked, fixes coming...')
def test_config_show_empty(runner, monkeypatch):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.config, ["show"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert "No configuration" in result.output

@pytest.mark.skip(reason='config reworked, fixes coming...')
def test_config_set(runner):
    with runner.isolated_filesystem():
        runner.invoke(cli.config, ["init"])
        result = runner.invoke(cli.config, ["set", "foo=bar"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert "foo='bar'" in result.output

@pytest.mark.skip(reason='config reworked, fixes coming...')
def test_config_del(runner):
    with runner.isolated_filesystem():
        with open("config-defaults.yaml", "w") as f:
            f.write(yaml.dump({'val': {'value': 'awesome', 'desc': 'cool'}, 'bad': {'value':'shit'}}))
        result = runner.invoke(cli.config, ["del", "bad"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert "1 parameters changed" in result.output

def test_push(runner, request_mocker, query_project, upload_url, upsert_run, monkeypatch):
    query_project(request_mocker)
    upload_url(request_mocker)
    update_mock = upsert_run(request_mocker)
    with runner.isolated_filesystem():
        #So GitRepo is in this cwd
        os.mkdir('wandb')
        monkeypatch.setattr(cli, 'api', Api({'project': 'test'}))
        with open("wandb/latest.yaml", "w") as f:
            f.write(yaml.dump({'wandb_version': 1, 'test': {'value': 'success', 'desc': 'My life'}}))
        with open('weights.h5', 'wb') as f:
            f.write(os.urandom(5000))
        result = runner.invoke(cli.push, ['test/default', 'weights.h5', '-m', 'My description'])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        assert "Updating run: test/default" in result.output

def test_push_no_run(runner):
    with runner.isolated_filesystem():
        with open('weights.h5', 'wb') as f:
            f.write(os.urandom(5000))
        result = runner.invoke(cli.push, ['weights.h5', '-p', 'test', '-m', 'Something great'])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 2
        assert "Run id is required if files are specified." in result.output

def test_push_dirty_git(runner, monkeypatch):
    with runner.isolated_filesystem():
        os.mkdir('wandb')

        # If the test was run from a directory containing .wandb, then __stage_dir__
        # was '.wandb' when imported by api.py, reload to fix. UGH!
        reload(wandb)
        repo = git_repo()
        open("foo.txt", "wb").close()
        repo.repo.index.add(["foo.txt"])
        monkeypatch.setattr(cli, 'api', Api({'project': 'test'}))
        cli.api._settings['git_tag'] = True
        result = runner.invoke(cli.push, ["test", "foo.txt", "-p", "test", "-m", "Dirty"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 1
        assert "You have un-committed changes." in result.output

def test_push_dirty_force_git(runner, request_mocker, query_project, upload_url, upsert_run, monkeypatch):
    query_project(request_mocker)
    upload_url(request_mocker)
    update_mock = upsert_run(request_mocker)
    with runner.isolated_filesystem():
        #So GitRepo is in this cwd
        monkeypatch.setattr(cli, 'api', Api({'project': 'test'}))
        repo = git_repo()
        with open('weights.h5', 'wb') as f:
            f.write(os.urandom(100))
        repo.repo.index.add(["weights.h5"])
        result = runner.invoke(cli.push, ["test", "weights.h5", "-f", "-p", "test", "-m", "Dirty"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0

def test_push_auto(runner, request_mocker, mocker, query_project, upload_url, monkeypatch):
    query_project(request_mocker)
    upload_url(request_mocker)
    edit_mock = mocker.patch("click.edit")
    with runner.isolated_filesystem():
        #So GitRepo is in this cwd
        monkeypatch.setattr(cli, 'api', Api({'project': 'test'}))
        with open('weights.h5', 'wb') as f:
            f.write(os.urandom(5000))
        with open('fake.json', 'wb') as f:
            f.write(os.urandom(100))
        result = runner.invoke(cli.push, ['test', '--project', 'test', '-m', 'Testing'])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        assert "Uploading file: weights.h5" in result.output
        #TODO: test without specifying message
        #assert edit_mock.called

def test_pull(runner, request_mocker, query_project, download_url):
    query_project(request_mocker)
    download_url(request_mocker)
    with runner.isolated_filesystem():
        result = runner.invoke(cli.pull, ['test', '--project', 'test'])

        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        assert "Downloading: test/test" in result.output
        assert os.path.isfile("weights.h5")
        assert "File model.json" in result.output
        assert "File weights.h5" in result.output

def test_pull_custom_run(runner, request_mocker, query_project, download_url):
    query_project(request_mocker)
    download_url(request_mocker)
    with runner.isolated_filesystem():
        result = runner.invoke(cli.pull, ['test/test'])

        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        assert "Downloading: test/test" in result.output

def test_pull_empty_run(runner, request_mocker, query_empty_project, download_url):
    query_empty_project(request_mocker)
    result = runner.invoke(cli.pull, ['test/test'])

    print(result.output)
    print(result.exception)
    print(traceback.print_tb(result.exc_info[2]))
    assert result.exit_code == 1
    assert "Run has no files" in result.output

def test_projects(runner, request_mocker, query_projects):
    query_projects(request_mocker)
    result = runner.invoke(cli.projects)
    assert result.exit_code == 0
    assert "test_2 - Test model" in result.output

def test_status(runner, request_mocker, query_project):
    with runner.isolated_filesystem():
        query_project(request_mocker)
        result = runner.invoke(cli.status, ["-p", "foo"])
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        assert "/latest" in result.output

def test_status_project_and_run(runner, request_mocker, query_project):
    query_project(request_mocker)
    result = runner.invoke(cli.status, ["test/awesome"])
    print(result.output)
    print(result.exception)
    print(traceback.print_tb(result.exc_info[2]))
    assert result.exit_code == 0
    assert "test/awesome" in result.output


def test_no_project_bad_command(runner):
    result = runner.invoke(cli.cli, ["fsd"])
    print(result.output)
    print(result.exception)
    print(traceback.print_tb(result.exc_info[2]))
    assert "No such command" in result.output
    assert result.exit_code == 2


@pytest.mark.skip('restore functionality broken for now, fixes coming...')
def test_restore(runner, request_mocker, query_run, monkeypatch):
    mock = query_run(request_mocker)
    with runner.isolated_filesystem():
        os.mkdir("wandb")
        repo = git_repo()
        with open("patch.txt", "w") as f:
            f.write("test")
        repo.repo.index.add(["patch.txt"])
        repo.repo.commit()
        monkeypatch.setattr(cli, 'api', Api({'project': 'test'}))
        result = runner.invoke(cli.restore, ["test/abcdef"])
        print(result.output)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        assert "Created branch wandb/abcdef" in result.output
        assert "Applied patch" in result.output
        assert "Restored config variables" in result.output

def test_projects_error(runner, request_mocker, query_projects):
    query_projects(request_mocker, status_code=400)
    result = runner.invoke(cli.projects)
    assert result.exit_code == 1
    print(result.output)
    assert "Error" in result.output

def test_init_new_login(runner, empty_netrc, local_netrc, request_mocker, query_projects, query_viewer):
    query_viewer(request_mocker)
    query_projects(request_mocker)
    with runner.isolated_filesystem():
        # If the test was run from a directory containing .wandb, then __stage_dir__
        # was '.wandb' when imported by api.py, reload to fix. UGH!
        reload(wandb)
        result = runner.invoke(cli.init, input="12345\nvanpelt")
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        with open("netrc", "r") as f:
            generatedNetrc = f.read()
        with open("wandb/settings", "r") as f:
            generatedWandb = f.read()
        assert "12345" in generatedNetrc
        assert "test_model" in generatedWandb

def test_init_reinit(runner, empty_netrc, local_netrc, request_mocker, query_projects, query_viewer):
    query_viewer(request_mocker)
    query_projects(request_mocker)
    with runner.isolated_filesystem():
        os.mkdir('wandb')
        result = runner.invoke(cli.init, input="12345\nvanpelt\ny")
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        with open("netrc", "r") as f:
            generatedNetrc = f.read()
        with open("wandb/settings", "r") as f:
            generatedWandb = f.read()
        assert "12345" in generatedNetrc
        assert "test_model" in generatedWandb

def test_init_add_login(runner, empty_netrc, local_netrc, request_mocker, query_projects, query_viewer):
    query_viewer(request_mocker)
    query_projects(request_mocker)
    with runner.isolated_filesystem():
        with open("netrc", "w") as f:
            f.write("previous config")
        result = runner.invoke(cli.init, input="12345\nvanpelt\n")
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        with open("netrc", "r") as f:
            generatedNetrc = f.read()
        with open("wandb/settings", "r") as f:
            generatedWandb = f.read()
        assert "12345" in generatedNetrc
        assert "previous config" in generatedNetrc

def test_init_existing_login(runner, local_netrc, request_mocker, query_projects, query_viewer):
    query_viewer(request_mocker)
    query_projects(request_mocker)
    with runner.isolated_filesystem():
        with open("netrc", "w") as f:
            f.write("machine api.wandb.ai\n\ttest\t12345")
        result = runner.invoke(cli.init, input="vanpelt\n")
        print(result.output)
        print(result.exception)
        print(traceback.print_tb(result.exc_info[2]))
        assert result.exit_code == 0
        with open("wandb/settings", "r") as f:
            generatedWandb = f.read()
        assert "test_model" in generatedWandb
        assert "This directory is configured" in result.output



# -*- coding: utf-8 -*-
import click
import sys
import copy
import random
import time
import os
import re
import netrc
import logging
import json
import glob
import io
import stat
import subprocess
from functools import wraps
from click.utils import LazyFile
from click.exceptions import BadParameter, ClickException
import inquirer
import sys
import traceback
import textwrap

from wandb import util
from wandb import Api, Error, Config, __version__, __stage_dir__
from wandb import wandb_run

logger = logging.getLogger(__name__)


def write_netrc(host, entity, key):
    """Add our host and key to .netrc"""
    try:
        normalized_host = host.split("/")[-1].split(":")[0]
        print("Appending to netrc %s" % os.path.expanduser('~/.netrc'))
        with open(os.path.expanduser('~/.netrc'), 'a') as f:
            f.write("""machine {host}
        login {entity}
        password {key}
    """.format(host=normalized_host, entity=entity, key=key))
        os.chmod(os.path.expanduser('~/.netrc'), stat.S_IRUSR | stat.S_IWUSR)
    except IOError as e:
        click.secho("Unable to read ~/.netrc", fg="red")
        return None


def display_error(func):
    """Function decorator for catching common errors and re-raising as wandb.Error"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Error as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            logger.error('\n'.join(lines))
            raise ClickException(e)

    return wrapper


def _require_init():
    if __stage_dir__ is None:
        print('Directory not initialized. Please run "wandb init" to get started.')
        sys.exit(1)


def editor(content='', marker='# Enter a description, markdown is allowed!\n'):
    message = click.edit(content + '\n\n' + marker)
    if message is not None:
        return message.split(marker, 1)[0].rstrip('\n')


api = Api()

# Some commands take project/entity etc. as arguments. We provide default
# values for those arguments from the current project configuration, as
# returned by api.settings()
CONTEXT = dict(default_map=api.settings())


class RunGroup(click.Group):
    @display_error
    def get_command(self, ctx, cmd_name):
        # TODO: check if cmd_name is a file in the current dir and not require `run`?
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        return None


@click.command(cls=RunGroup)
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx):
    """Weights & Biases

If the first argument is a file in the current directory run it.

   wandb train.py --arg=1
    """
    pass


@cli.command(context_settings=CONTEXT, help="List projects")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def projects(entity, display=True):
    projects = api.list_projects(entity=entity)
    if len(projects) == 0:
        message = "No projects found for %s" % entity
    else:
        message = 'Latest projects for "%s"' % entity
    if display:
        click.echo(click.style(message, bold=True))
        for project in projects:
            click.echo("".join(
                (click.style(project['name'], fg="blue", bold=True),
                 " - ",
                 str(project['description']).split("\n")[0])
            ))
    return projects


@cli.command(context_settings=CONTEXT, help="List runs in a project")
@click.option("--project", "-p", prompt=True, envvar='WANDB_PROJECT', help="The project you wish to list runs from.")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def runs(project, entity):
    click.echo(click.style('Latest runs for project "%s"' %
                           project, bold=True))
    runs = api.list_runs(project, entity=entity)
    for run in runs:
        click.echo("".join(
            (click.style(run['name'], fg="blue", bold=True),
             " - ",
             (run['description'] or "").split("\n")[0])
        ))


@cli.command(context_settings=CONTEXT, help="List local & remote file status")
@click.argument("run", envvar='WANDB_RUN')
@click.option("--settings/--no-settings", help="Show the current settings", default=False)
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@display_error
def status(run, settings, project):
    if settings:
        click.echo(click.style("Current Settings", bold=True) +
                   " (%s)" % api.settings_file)
        settings = api.settings()
        click.echo(json.dumps(
            settings,
            sort_keys=True,
            indent=2,
            separators=(',', ': ')
        ))
        click.echo(click.style("Logged in?", bold=True) + " %s\n" %
                   bool(api.api_key))
    project, run = api.parse_slug(run, project=project)
    existing = set()  # TODO: populate this set with the current files in the run dir
    remote = api.download_urls(project, run)
    not_synced = set()
    remote_names = set([name for name in remote])
    for file in existing:
        meta = remote.get(file)
        if meta and not api.file_current(file, meta['md5']):
            not_synced.add(file)
        elif not meta:
            not_synced.add(file)
    # TODO: remove items that exists and have the md5
    only_remote = remote_names.difference(existing)
    up_to_date = existing.difference(only_remote).difference(not_synced)
    click.echo('File status for ' + click.style('"%s/%s" ' %
                                                (project, run), bold=True))
    if len(not_synced) > 0:
        click.echo(click.style('Push needed: ', bold=True) +
                   click.style(", ".join(not_synced), fg="red"))
    if len(only_remote) > 0:
        click.echo(click.style('Pull needed: ', bold=True) +
                   click.style(", ".join(only_remote), fg="red"))
    if len(up_to_date) > 0:
        click.echo(click.style('Up to date: ', bold=True) +
                   click.style(", ".join(up_to_date), fg="green"))


@cli.command(context_settings=CONTEXT, help="Store notes for a future training run")
@display_error
def describe():
    path = __stage_dir__ + 'description.md'
    existing = (os.path.exists(path) and open(path).read()) or ''
    description = editor(existing)
    if description:
        with open(path, 'w') as file:
            file.write(description)
    click.echo(
        "Notes stored for next training run\nCalling wandb.sync() in your training script will persist them.")


@cli.command(context_settings=CONTEXT, help="Restore code and config state for a run")
@click.argument("run", envvar='WANDB_RUN')
@click.option("--branch/--no-branch", default=True, help="Whether to create a branch or checkout detached")
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def restore(run, branch, project, entity):
    project, run = api.parse_slug(run, project=project)
    commit, json_config, patch = api.run_config(
        project, run=run, entity=entity)
    if commit:
        branch_name = "wandb/%s" % run
        if branch and branch_name not in api.git.repo.branches:
            api.git.repo.git.checkout(commit, b=branch_name)
            click.echo("Created branch %s" %
                       click.style(branch_name, bold=True))
        elif branch:
            click.secho(
                "Using existing branch, run `git branch -D %s` from master for a clean checkout" % branch_name, fg="red")
            api.git.repo.git.checkout(branch_name)
        else:
            click.secho("Checking out %s in detached mode" % commit)
            api.git.repo.git.checkout(commit)

    if patch:
        with open(__stage_dir__ + "diff.patch", "w") as f:
            f.write(patch)
        api.git.repo.git.execute(
            ['git', 'apply', __stage_dir__ + 'diff.patch'])
        click.echo("Applied patch")

    config = Config()
    config.load_json(json_config)
    config.persist()
    click.echo("Restored config variables")


@cli.command(context_settings=CONTEXT, help="Push files to Weights & Biases")
@click.argument("run", envvar='WANDB_RUN')
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you wish to upload to.")
@click.option("--description", "-m", help="A description to associate with this upload.")
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@click.option("--force/--no-force", "-f", default=False, help="Whether to force git tag creation.")
@click.argument("files", type=click.File('rb'), nargs=-1)
@click.pass_context
@display_error
def push(ctx, run, project, description, entity, force, files):
    # TODO: do we support the case of a run with the same name as a file?
    if os.path.exists(run):
        raise BadParameter("Run id is required if files are specified.")
    project, run = api.parse_slug(run, project=project)

    click.echo("Updating run: {project}/{run}".format(
        project=click.style(project, bold=True), run=run))

    candidates = []
    if len(files) == 0:
        # TODO: do we want to do this?
        patterns = ("*.h5", "*.hdf5", "*.json", "*.meta", "*checkpoint*")
        for pattern in patterns:
            candidates.extend(glob.glob(pattern))
        if len(candidates) == 0:
            raise BadParameter(
                "Couldn't auto-detect files, specify manually or use `wandb.add`", param_hint="FILES")

        choices = inquirer.prompt([inquirer.Checkbox('files', message="Which files do you want to push? (left and right arrows to select)",
                                                     choices=[c for c in candidates])])
        files = [LazyFile(choice, 'rb') for choice in choices['files']]

    # TODO: Deal with files in a sub directory
    api.push(project, files=[f.name for f in files], run=run,
             description=description, entity=entity, force=force, progress=sys.stdout)


@cli.command(context_settings=CONTEXT, help="Pull files from Weights & Biases")
@click.argument("run", envvar='WANDB_RUN')
@click.option("--project", "-p", envvar='WANDB_PROJECT', help="The project you want to download.")
@click.option("--kind", "-k", default="all", type=click.Choice(['all', 'model', 'weights', 'other']))
@click.option("--entity", "-e", default="models", envvar='WANDB_ENTITY', help="The entity to scope the listing to.")
@display_error
def pull(project, run, kind, entity):
    project, run = api.parse_slug(run, project=project)

    urls = api.download_urls(project, run=run, entity=entity)
    if len(urls) == 0:
        raise ClickException("Run has no files")
    click.echo("Downloading: {project}/{run}".format(
        project=click.style(project, bold=True), run=run
    ))

    for name in urls:
        if api.file_current(name, urls[name]['md5']):
            click.echo("File %s is up to date" % name)
        else:
            length, response = api.download_file(urls[name]['url'])
            with click.progressbar(length=length, label='File %s' % name,
                                   fill_char=click.style('&', fg='green')) as bar:
                with open(name, "wb") as f:
                    for data in response.iter_content(chunk_size=4096):
                        f.write(data)
                        bar.update(len(data))


@cli.command(context_settings=CONTEXT, help="Login to Weights & Biases")
@display_error
def login():
    # Import in here for performance reasons
    import webbrowser
    # TODO: use Oauth and a local webserver: https://community.auth0.com/questions/6501/authenticating-an-installed-cli-with-oidc-and-a-th
    url = "https://app.wandb.ai/profile"
    # TODO: google cloud SDK check_browser.py
    launched = webbrowser.open_new_tab(url)
    if launched:
        click.echo(
            'Opening [{0}] in a new tab in your default browser.'.format(url))
    else:
        click.echo("You can find your API keys here: {0}".format(url))

    key = click.prompt("{warning} Paste an API key from your profile".format(
        warning=click.style("Not authenticated!", fg="red")),
        value_proc=lambda x: x.strip())

    host = api.settings()['base_url']
    if key:
        # TODO: get the username here...
        # username = api.viewer().get('entity', 'models')
        write_netrc(host, "user", key)


@cli.command(context_settings=CONTEXT, help="Configure a directory with Weights & Biases")
@click.pass_context
@display_error
def init(ctx):
    # TODO: This is commented out because we always automatically create this dir in __init__.py
    # however this isn't ideal, we'll litter the filesystem with wandb directories.
    # if(os.path.exists(__stage_dir__)):
    #    click.confirm(click.style("This directory is already configured, should we overwrite it?", fg="red"), abort=True)
    click.echo(click.style(
        "Let's setup this directory for W&B!", fg="green", bold=True))
    global api

    if api.api_key is None:
        ctx.invoke(login)
        api = Api()

    entity = click.prompt("What username or org should we use?",
                          default=api.viewer().get('entity', 'models'))
    # TODO: handle the case of a missing entity
    result = ctx.invoke(projects, entity=entity, display=False)

    if len(result) == 0:
        project = click.prompt("Enter a name for your first project")
        description = editor()
        api.upsert_project(project, entity=entity, description=description)
    else:
        project_names = [project["name"] for project in result]
        question = inquirer.List(
            'project', message="Which project should we use?", choices=project_names + ["Create New"])
        project = inquirer.prompt([question])['project']
        # TODO: check with the server if the project exists
        if project == "Create New":
            project = click.prompt("Enter a name for your new project")
            description = editor()
            api.upsert_project(project, entity=entity, description=description)
        else:
            ids = [res['id'] for res in result if res['name'] == project]
            if len(ids) > 0:
                api.upsert_project(project, id=ids[0], entity=entity)

    from wandb import _set_stage_dir, get_stage_dir
    if get_stage_dir() is None:
        _set_stage_dir('wandb')

    from wandb import get_stage_dir
    wandb_path = os.path.join(os.getcwd(), get_stage_dir())
    if os.path.isdir(wandb_path):
        click.confirm(click.style(
            "This directory is already initialized, should we overwrite it?", fg="red"), abort=True)
    else:
        os.mkdir(wandb_path)

    with open(os.path.join(get_stage_dir(), 'settings'), "w") as file:
        file.write("[default]\nentity: {entity}\nproject: {project}\n".format(
            entity=entity, project=project))

    with open(os.path.join(get_stage_dir(), '.gitignore'), "w") as file:
        file.write("*\n!settings")

    with open('config-defaults.yaml', 'w') as file:
        file.write(textwrap.dedent("""\
            wandb_version: 1

            # Example variables below. Uncomment (remove leading '# ') to use them, or just
            # delete and create your own.
            
            # epochs:
            #   desc: Number of epochs to train over
            #   value: 100
            # batch_size:
            #   desc: Size of each mini-batch
            #   value: 32
            """))

    click.echo(click.style("This directory is configured!  Try these next:\n", fg="green") +
               """
* Track runs: `{code}` in your training script, then `{run}`.
* Run `{push}` to manually add a file.
* Pull popular models into your project with: `{pull}`.
    """.format(
        push=click.style("wandb push run_id weights.h5", bold=True),
        code=click.style("import wandb", bold=True),
        run=click.style("wandb run <training_command>", bold=True),
        pull=click.style("wandb pull models/inception-v4", bold=True)
    ))


RUN_CONTEXT = copy.copy(CONTEXT)
RUN_CONTEXT['allow_extra_args'] = True
RUN_CONTEXT['ignore_unknown_options'] = True


@cli.command(context_settings=RUN_CONTEXT, help="Launch a job")
@click.pass_context
@click.argument('program')
@click.argument('args', nargs=-1)
@click.option('--id', default=None,
              help='Run id to use, default is to generate.')
@click.option('--dir', default=None,
              help='Files in this directory will be saved to wandb, defaults to wandb/run-<run_id>')
@click.option('--configs', default=None,
              help='Config file paths to load')
@display_error
def run(ctx, program, args, id, dir, configs):
    _require_init()
    env = copy.copy(os.environ)
    env['WANDB_MODE'] = 'run'
    if id is None:
        id = wandb_run.generate_id()
    env['WANDB_RUN_ID'] = id
    if dir is None:
        dir = wandb_run.run_dir_path(id, dry=False)
        util.mkdir_exists_ok(dir)
    env['WANDB_RUN_DIR'] = dir
    if configs:
        env['WANDB_CONFIG_PATHS'] = configs
    command = [program] + list(args)
    runner = util.find_runner(program)
    if runner:
        command = runner.split() + command
    proc = util.SafeSubprocess(command, env=env, read_output=False)
    proc.run()
    while True:
        time.sleep(0.1)
        exitcode = proc.poll()
        if exitcode is not None:
            print('wandb: job (%s) Process exited with code: %s' %
                  (program, exitcode))
            break


#@cli.group()
#@click.pass_context
#@display_error
# def config(ctx):
#    """Manage this projects configuration.
#
# Examples:
#
#    wandb config set param=2 --description="Some tunning parameter"
#    wandb config del param
#    wandb config show
#    """
#    pass


#@config.command("init", help="Initialize a directory with wandb configuration")
#@display_error
# def config_init(prompt=True):
#    from wandb import get_stage_dir
#    config_path = os.path.join(os.getcwd(), get_stage_dir())
#    config = Config()
#    if os.path.isdir(config_path):
#        if prompt:
#            click.confirm(click.style("This directory is already initialized, should we overwrite it?", fg="red"), abort=True)
#    else:
#        os.mkdir(config_path)
#    config.epochs_desc = "Number epochs to train over"
#    config.epochs = 32
#    config.persist()
#    if prompt:
#        click.echo("""Configuration initialized, use `wandb config set` to set parameters.  Then in your training script:
# import wandb
# conf = wandb.sync()
# conf.batch_size
#""")


#@config.command(help="Show the current config")
#@click.option("--format", help="The format to dump the config as", default="python", type=click.Choice(['python', 'yaml', 'json']))
#@display_error
# def show(format, changed=[], diff=False):
#    if len(changed) == 0 and diff:
#        click.secho("No parameters were changed", fg="red")
#    elif diff:
#        click.echo("%i parameters changed: " % len(changed))
#    config = Config()
#    if len(vars(config)) == 0:
#        click.secho(
#            "No configuration found in this directory, run `wandb config init`", fg="red")
#    if format == "yaml":
#        click.echo("%s" % config)
#    elif format == "json":
#        click.echo(json.dumps(vars(config)))
#    elif format == "python":
#        res = ""
#        for key in set(config.keys + changed):
#            if config.desc(key):
#                res += "# %s\n" % config.desc(key)
#            style = None
#            if key in changed:
#                style = "green" if config.get(key) else "red"
#            res += click.style("%s=%r\n" % (key, config.get(key)),
#                               bold=True if style is None else False, fg=style)
#        click.echo(res)


#@config.command("import", help="Import configuration parameters")
#@click.option("--format", "-f", help="The format to parse the imported params", default="python", type=click.Choice(["python"]))
#@click.pass_context
#@display_error
# def import_config(ctx, format):
#    data = editor("# Paste python comments and variable definitions above")
#    desc = None
#    config = Config()
#    imported = []
#    if data:
#        for line in data.split("\n"):
#            if line.strip().startswith("#"):
#                desc = line.strip(" #")
#            elif "=" in line:
#                try:
#                    key, value = [str(part.strip())
#                                  for part in line.split("=")]
#                    if len(value) == 0:
#                        continue
#                    config[key] = value
#                    imported.append(key)
#                    if desc:
#                        config[key + "_desc"] = desc
#                    desc = None
#                except ValueError:
#                    logging.error("Invalid line: %s" % line)
#            else:
#                logging.warn("Skipping line %s", line)
#        config.persist()
#    ctx.invoke(show, changed=imported, diff=True)


#@config.command("set", help="Set config variables with key=value pairs")
#@click.argument("key_values", nargs=-1)
#@click.option("--description", "-d", help="A description for the config value if specifying one pair")
#@click.pass_context
#@display_error
# def config_set(ctx, key_values, description=None):
#    config = Config()
#    if len(key_values) == 0:
#        raise ClickException(
#            "Must specify at least 1 key value pair i.e. `wandb config set epochs=11`")
#    if len(key_values) > 1 and description:
#        raise ClickException(
#            "Description can only be specified with 1 key value pair.")
#    changed = []
#    for pair in key_values:
#        try:
#            key, value = pair.split("=")
#        except ValueError:
#            key = pair
#            value = None
#        if value:
#            changed.append(key)
#            config[str(key)] = value
#        if description:
#            config[str(key) + "_desc"] = description
#    config.persist()
#    ctx.invoke(show, changed=changed, diff=True)


#@config.command("del", help="Delete config variables")
#@click.argument("keys", nargs=-1)
#@click.pass_context
#@display_error
# def delete(ctx, keys):
#    config = Config()
#    if len(keys) == 0:
#        raise ClickException(
#            "Must specify at least 1 key i.e. `wandb config rm epochs`")
#    changed = []
#    for key in keys:
#        del config[str(key)]
#        changed.append(key)
#    config.persist()
#    ctx.invoke(show, changed=changed, diff=True)


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-

__author__ = """Chris Van Pelt"""
__email__ = 'vanpelt@wandb.com'
__version__ = '0.4.22'

import types
import six
import sys
import logging
import os
import traceback

# We use the hidden version if it already exists, otherwise non-hidden.
if os.path.exists('.wandb'):
    __stage_dir__ = '.wandb/'
elif os.path.exists('wandb'):
    __stage_dir__ = "wandb/"
else:
    __stage_dir__ = None


def get_stage_dir():
    return __stage_dir__

# Used when initing a new project with "wandb init"


def _set_stage_dir(stage_dir):
    global __stage_dir__
    __stage_dir__ = stage_dir


from .git_repo import GitRepo
from .api import Api, Error
from .sync import Sync
from .config import Config
from .results import Results
from .summary import Summary
from .history import History
from wandb import wandb_run

# Three possible modes:
#     'cli': running from "wandb" command
#     'run': we're a script launched by "wandb run"
#     'dryrun': we're a script not launched by "wandb run"

# Hmmm....
stack_frame0 = traceback.extract_stack()[0]
try:
    launched_file = stack_frame0.filename
except AttributeError:
    launched_file = stack_frame0[0]

if launched_file.endswith('bin/wandb'):
    MODE = 'cli'
else:
    MODE = os.environ.get('WANDB_MODE', 'dryrun')


# called by cli.py
# Even when running the wandb cli, __init__.py is imported before main() runs, so we set
# cli mode afterward. This means there's a period of time before this call when MODE will
# be dryrun
def _set_cli_mode():
    global MODE, run
    MODE = 'cli'
    run = None


if __stage_dir__ is not None:
    log_fname = __stage_dir__ + 'debug.log'
else:
    log_fname = './wandb-debug.log'
logging.basicConfig(
    filemode="w",
    filename=log_fname,
    level=logging.DEBUG)


def push(*args, **kwargs):
    Api().push(*args, **kwargs)


def pull(*args, **kwargs):
    Api().pull(*args, **kwargs)


def _do_sync(dir, extra_config=None):
    if MODE == 'run':
        api = Api()
        if api.api_key is None:
            raise Error(
                "No API key found, run `wandb login` or set WANDB_API_KEY")
        api.set_current_run_id(run.id)
        if extra_config is not None:
            run.config.update(extra_config)
        sync = Sync(api, run.id, config=run.config, dir=dir)
        sync.watch(files='*')
        return sync
    elif MODE == 'dryrun':
        print('wandb dryrun mode. Use "wandb run <script>" to save results to wandb.\n'
              'Run directory: %s' % run.dir)


# The current run (a Run object)
run = None

if __stage_dir__ and MODE != 'cli':
    _run_id = os.getenv('WANDB_RUN_ID')
    if _run_id is None:
        _run_id = wandb_run.generate_id()
    _run_dir = os.getenv('WANDB_RUN_DIR')
    if _run_dir is None:
        _run_dir = wandb_run.run_dir_path(_run_id, dry=MODE == 'dryrun')
    _conf_paths = os.getenv('WANDB_CONFIG_PATHS', '')
    if _conf_paths:
        _conf_paths = _conf_paths.split(',')

    syncer = None

    def persist_config_callback():
        if syncer:
            syncer.update_config(_config)
    _config = Config(config_paths=_conf_paths,
                     wandb_dir=__stage_dir__, run_dir=_run_dir,
                     persist_callback=persist_config_callback)
    run = wandb_run.Run(_run_id, _run_dir, _config)
    syncer = _do_sync(run.dir)

__all__ = ["Api", "Error", "Config", "Results", "History", "Summary",
           "WandBKerasCallback"]
